/*
        download.h

        Pacsat Broadcast protocol receiver

        Bent Bagger
        OZ6BL

        Bregnerødvej 151
        3460 Birkerød
        Birkerød
        Denmark

        INTERNET:       oz6bl@amsat.org
*/

#define DOWNLOAD_H_ID "$Id: download.h,v 1.2 1999/12/15 20:07:58 bent Exp $"
      
typedef struct LOADLIST_STR
{
        struct LOADLIST_STR     *next;
        char                    dest  [10];
        char                    src   [10];
        char                    title [40];
        char                    key   [40];
        int                     size;
}
LOADLIST;
 
