/*
#define RCS_ID $Id: request.h,v 1.2 1999/12/15 20:25:42 bent Exp $
	request.h

	Defines for inter-program communications.


	John Melton
	G0ORX, N6LYT

	4 Charlwoods Close
	Copthorne
	West Sussex
	RH10 3QZ
	England

	INTERNET:	g0orx@amsat.org
			n6lyt@amsat.org
			john@images.demon.co.uk
			J.D.Melton@slh0613.icl.wins.co.uk

	Extended by:
	Peter Scott
	oz2aba, oz7sat
*/

struct request
{
	int type;
#define	REQ_TYPE_CANCEL	0
#define	REQ_TYPE_DIR	1
#define	REQ_TYPE_FILE	2
#define REQ_TYPE_REFILE	3
#define REQ_TYPE_QUIT	4
#define REQ_TYPE_REINIT 5

	unsigned long fileId;
};

