/*
	xpb.h

	eXperimental Pacsat Broadcast protocol receiver

	John Melton
	G0ORX, N6LYT

	4 Charlwoods Close
	Copthorne
	West Sussex
	RH10 3QZ
	England

	INTERNET:	g0orx@amsat.org
			n6lyt@amsat.org
			john@images.demon.co.uk
			J.D.Melton@slh0613.icl.wins.co.uk
*/

#define	MAXBUFFER	512
/* #define	MAXFILES	20		/ * G4KLX from 8 */
#define	MAXFILES 10

#define MAXDOWNLOAD	1000		/* Size of download.req file */

typedef struct HOLE_STR
{
	unsigned	start;
	unsigned	finish;
	struct HOLE_STR	*nextHole;
}
HOLE;

typedef struct
{
	unsigned	fileId;
	FILE		*file;
	int		hdrSeen;
	int		fileSize;
	int		holes;
	HOLE		*firstHole;
}
OPENFILES;
